/*
SPLICE 1.5 : a C++ utility for "gene" splicing, cutting, copying, inserting and
        overlaying bits of a file into another
        Uses for SPLICE:
        1. excising a section of a file.
        2. replacing a section of a file.
        3. inserting a section into a file.
        4. extracting a section of a file.
        5. appending a section onto a file.
        This program may be copied freely and used for any purpose except military.
        */
/*
Roedy Green
Canadian Mind Products
#101 - 2536 Wark Street
Victoria BC Canada
V8T 4G8
tel:(250) 361-9093
http://mindprod.com
*/
/*
A windows console app.
For Visual C++ Express options see http://mindprod.com/jgloss/cpp.html#COMPILING
SPLICE commands, may be upper or lower case:
CUT MyFile1.Txt START 6 LENGTH 5 RECSIZE 1024
    remove a slice from a file offset 6 length 5
    Offsets and lengths are measured in blocks of 1024.
CUT MyFile1.Txt START 6 LENGTH 5 TO MyFile2.txt RECSIZE 1024
    remove a slice from a file offset 6 length 5
    and copy it to a new file.
CUT MyFile1.Txt START 6 LENGTH 5 APPEND MyFile2.txt RECSIZE 1024
    remove a slice from a file offset 6 length 5
    and copy it to the end of another file.
CUT MyFile1.Txt START 6 LENGTH 5 OVERLAY MyFile2.txt AT 7 RECSIZE 1024
    remove a slice from a file, offset 6 length 5,
    overlay it on top of another file at offset 7.
CUT MyFile1.Txt START 6 LENGTH 5 INSERT MyFile2.txt AT 7 RECSIZE 1024
    remove a slice from a file, offset 6, length 5, insert
    at offset 7 in another file
COPY MyFile1.Txt START 6 LENGTH 5 TO MyFile2.txt RECSIZE 1024
    copy a slice from a file offset 6 length 5
    to a new file
COPY MyFile1.Txt START 6 LENGTH 5 APPEND MyFile2.txt RECSIZE 1024
    copy a slice from a file offset 6 length 5
    to the end of another file.
COPY MyFile1.Txt START 6 LENGTH 5 OVERLAY MyFile2.txt AT 7 RECSIZE 1024
    copy a slice from a file, offset 6 length 5,
    overlaying it on top of another file at offset 7,
    replacing the previous data.
COPY MyFile1.Txt START 6 LENGTH 5 INSERT MyFile2.txt AT 7 RECSIZE 1024
    copy a slice from a file, offset 6, length 5, inserting
    at offset 7 in another file.
START defaults to 0
AT defaults to 0
RECSIZE defaults to 1
If length reaches out past the end of the source file, it will be shortened.
Target files will be expanded as needed with zeroes.
Version History:
 1.0 1997-04-27 initial release
 1.1 1997-04-28 added APPEND
                add abbreviations for keywords
 1.2 1998-11-08 embed Barker address
                correct Bulk::move logic so that fromOffset is defined.
 1.3 2007-01-04 convert to POSIX, safe char handling lib, long filename support.
 1.4 2009-03-11 add ant build script and Visual C++ project build file.
 1.5 2010-11-07 fix delete[] bug the compiler did not detect.
*/
/* I N C L U D E S */
#include <assert.h> /* assert */
#include <stdlib.h>  /* exit, putenv, _splitpath */
#include <stdio.h>   /* fclose, fgetc, printf, remove, rename, setvbuf, FILE, printf, fopen */
#include <fcntl.h>
// O_CREAT O_EXCL
#include <sys\stat.h>
// S_IWRITE
#include <io.h>
// open, write
#include <share.h>
// _SH_DENYRW
#include <conio.h>   /* getch, putch */
// e.g. getch kbhit crlscr
#include <string.h>
// strcpy
#define complaintSize 256
/* P R O T O T Y P E S */
int main( int argc,  char *argv[] );
void oops( char* reason );
void banner();
// =============================================================================
class SeqFile
   {
public:
   // size of file in bytes
   int size;
   int handle;
   // we use DOS level handle open rather than fopen to get access to a way
   // to shrink files.
   // constructor
   SeqFile( char* internalName, char* fileName, int openFlags, int shareFlags, int permissionMode );
   // destructor
   ~SeqFile();
   }; // end SeqFile
// =============================================================================
class Bulk
   {
   // bulk moving data around in and between files
   // buffer size is 31K.  This makes it a 16-bit signed value.
#define bufferSize 31744
public:
   // clear area to constant
   static void zap( int theFile, int start, int length, unsigned char value );
   // move data within a file, either up or down
   static void move( int theFile, int start, int length, int at );
   // move data between two different files
   static void move( int fromFile, int start, int length, int toFile, int at );
   static void resize( int theFile, int size );
   }; // end Bulk
// =============================================================================
typedef enum SourceType
   {
   unknown, cut, copy
   } SourceType;
typedef enum TargetType
   {
   none, to, append, overlay, insert
   } TargetType;
class Splice
   {
public:
   static void splice( int argc,  char *argv[] );
protected:
   // files
   static SeqFile* sourceFile;
   static SeqFile* targetFile;
   // offset control
   static int  start;
   static int  length;
   static int  at;
   static int  recsize;
   // source/target kind
   static SourceType source;
   static TargetType target;
   static void parse( int argc,  char *argv[] );
   static void checkParms();
   static void openFiles();
   static void analyse();
   static void work();
   static void closeFiles();
   // parms from the command line
   static char* cutName;
   static char* copyName;
   static char* toName;
   static char* appendName;
   static char* overlayName;
   static char* insertName;
   }; // end class Splice
// =============================================================================
/* S T A T I C   D E F I N I T I O N S   A N D   I N I T I A L I S A T I O N */
// offset control
int  Splice::start    = 0;
int  Splice::length   = 0;
int  Splice::at       = 0;
int  Splice::recsize  = 1;
// source/target
SourceType Splice::source = unknown;
TargetType Splice::target = none;
SeqFile* Splice::sourceFile = NULL;
SeqFile* Splice::targetFile = NULL;
// parms from the command line
char* Splice::cutName       = NULL;
char* Splice::copyName      = NULL;
char* Splice::toName        = NULL;
char* Splice::appendName    = NULL;
char* Splice::overlayName   = NULL;
char* Splice::insertName    = NULL;
// =============================================================================
/* M E T H O D   D E F I N I T I O N S */
SeqFile::SeqFile ( char* internalName, char* fileName, int openFlags, int shareFlags, int permissionMode )
   { // constructor
   size = 0;
   handle = -1;
   // non-zero return means error
   // see http://msdn2.microsoft.com/en-us/library/w64k0ytk(VS.80).aspx for meaning of flags.
   if ( fileName &&  _sopen_s(&handle, fileName, openFlags, shareFlags, permissionMode) )
      {
      char* complaint = new char[256];
      complaint[0] = 0;
      strcat_s(complaint, complaintSize, "Can't open ");
      strcat_s(complaint, complaintSize, internalName);
      strcat_s(complaint, complaintSize, " file ");
      strcat_s(complaint, complaintSize, fileName);
      strcat_s(complaint, complaintSize, ".");
      oops(complaint);  // won't return
      delete[] complaint;
      }
   if ( handle != -1 ) size = _filelength(handle);
   } // end SeqFile constructor
SeqFile::~SeqFile ()
   {
   // destructor
   if ( handle ) _close(handle);
   } // end SeqFile destructor
// =============================================================================
void Bulk::zap( int theFile, int start, int length, unsigned char value )
   {
   // clear area of file to a value
   if ( length == 0 ) return;
   unsigned char* buffer = new unsigned char [bufferSize];
   if ( buffer == NULL ) oops ("Not enough RAM.");
   for ( int i=0; i<bufferSize; i++ )
      {
      buffer[i] = value;
      }
   _lseek (theFile, start, SEEK_SET);
   // do big chunks first
   int end = start + ((length / bufferSize) * bufferSize);
   for ( int offset=start; offset<end; offset+=bufferSize )
      {
      assert( buffer );
      _write(theFile, buffer, bufferSize);
      } // end for
   // do the remainder
   int remainder = length % bufferSize;
   if ( remainder )
      {
      _write(theFile, buffer, remainder);
      }
   delete[] buffer;
   } // end zap
void Bulk::move( int theFile, int start, int length, int at )
   {
   // move data within a file, either up or down
   if ( start == at || length == 0 ) return;
   unsigned char* buffer = new unsigned char [bufferSize];
   if ( buffer == NULL ) oops ("Not enough RAM.");
   if ( start < at && at < start + length )
      {
      // because of overlap, work down, right to left
      int begin = start + length - bufferSize;
      int end = begin - ((length / bufferSize) * bufferSize);
      int toOffset = at + length - bufferSize;
      int fromOffset;
      for ( fromOffset=begin; fromOffset>end; fromOffset-=bufferSize )
         {
         assert( buffer );
         _lseek (theFile, fromOffset, SEEK_SET);
         _read (theFile, buffer, bufferSize);
         _lseek (theFile, toOffset, SEEK_SET);
         _write(theFile, buffer, bufferSize);
         toOffset -= bufferSize;
         }
      // do the remainder
      int remainder = length % bufferSize;
      if ( remainder )
         {
         _lseek (theFile, start, SEEK_SET);
         _read (theFile, buffer, remainder);
         _lseek (theFile, at, SEEK_SET);
         _write(theFile, buffer, remainder);
         }
      }
   else
      {
      // work up, left to right.
      // end is offset just past last big chunk
      int end = start + ((length / bufferSize) * bufferSize);
      int toOffset = at;
      int fromOffset;
      for ( fromOffset=start; fromOffset<end; fromOffset+=bufferSize )
         {
         assert( buffer );
         _lseek (theFile, fromOffset, SEEK_SET);
         _read (theFile, buffer, bufferSize);
         _lseek (theFile, toOffset, SEEK_SET);
         _write(theFile, buffer, bufferSize);
         toOffset += bufferSize;
         }
      // do the remainder
      int remainder = length % bufferSize;
      if ( remainder )
         {
         _lseek (theFile, fromOffset, SEEK_SET);
         _read (theFile, buffer, remainder);
         _lseek (theFile, toOffset, SEEK_SET);
         _write(theFile, buffer, remainder);
         }
      }
   delete[] buffer;
   } // end movedata
void Bulk::move( int fromFile, int start, int length, int toFile, int at )
   {
   // move data between two different files
   if ( length == 0 ) return;
   unsigned char* buffer = new unsigned char [bufferSize];
   if ( buffer == NULL ) oops ("Not enough RAM.");
   _lseek (fromFile, start, SEEK_SET);
   _lseek (toFile, at, SEEK_SET);
   // do big chunks first
   int end = start + ((length / bufferSize) * bufferSize);
   for ( int offset=start; offset<end; offset+=bufferSize )
      {
      assert( buffer );
      _read (fromFile, buffer, bufferSize);
      _write(toFile, buffer, bufferSize);
      } // end for
   // do the remainder
   int remainder =  length % bufferSize;
   if ( remainder )
      {
      _read (fromFile, buffer, remainder);
      _write(toFile, buffer, remainder);
      }
   delete[] buffer;
   } // end movedata
void Bulk::resize ( int theFile, int size )
   {
   // defeat the auto zero-padding feature on growth
   // since we will be filling the growth area ourselves.
   int oldsize = _filelength(theFile);
   unsigned char* buffer = new unsigned char [1];
   if ( buffer == NULL ) oops ("buffer allocation failed.");
   buffer[0] = 0;
   if ( oldsize + 1 < size )
      {
      _lseek( theFile, size-1, SEEK_SET );
      _write( theFile, buffer, 1 ); // write one null byte to grow the file
      }
   // may be shrinking or growing
   if ( _chsize(theFile, size) == -1 ) oops ("insufficient disk space");
 delete[] buffer;
   } // end resize
// =============================================================================
void Splice::checkParms()
   {
   char * sourceName = NULL;
   source = unknown;
   int counter = 0;
   if ( cutName )
      {
      source = cut;  sourceName = cutName;  counter++;
      }
   if ( copyName )
      {
      source = copy; sourceName = copyName; counter++;
      }
   if ( counter != 1 )
      oops("You must have either a CUT or COPY source.");
   char * targetName = NULL;
   target = none;
   counter = 0;
   if ( toName )
      {
      target = to;      targetName = toName;      counter++;
      }
   if ( appendName )
      {
      target = append;  targetName = appendName;  counter++;
      }
   if ( overlayName )
      {
      target = overlay; targetName = overlayName; counter++;
      }
   if ( insertName )
      {
      target = insert;  targetName = insertName;  counter++;
      }
   if ( !(counter == 1  || (counter == 0 && source == cut)) )
      oops ("You must have either a TO, OVERLAY or INSERT target.");
   if ( _stricmp(sourceName, targetName) == 0 )
      oops ("You can't have the source and target files the same.");
   if ( target == to && at )
      oops ("TO and AT are incompatible. Use OVERLAY or INSERT");
   if ( target == append && at )
      oops ("APPEND and AT are incompatible. Use OVERLAY or INSERT");
   if ( source == cut && target == none && at )
      oops ("Use START, not AT to CUT");
   if ( start < 0 )
      oops ("START must be >= 0");
   if ( length  <= 0 )
      oops ("LENGTH must be > 0");
   if ( at < 0 )
      oops ("AT must be >= 0");
   if ( recsize <= 0 )
      oops ("RECSIZE must be > 0");
   } // end checkParms
void Splice::parse ( int argc,  char *argv[] )
   {
   // parse the command line
   char*  keyword;
   char*  value;
   for ( int i=1; i<argc; i+=2 )
      {
      keyword = argv[i];
      if ( i+1 >= argc ) oops("Must have keyword/value pairs separarated by space");
      value = argv[i+1];
      if ( _stricmp(keyword,"append")      ==0 ) appendName = value;
      else if ( _stricmp(keyword,"app")    ==0 ) appendName = value;
      else if ( _stricmp(keyword,"at")     ==0 ) at        = atol(value);
      else if ( _stricmp(keyword,"copy")   ==0 ) copyName  = value;
      else if ( _stricmp(keyword,"cut")    ==0 ) cutName   = value;
      else if ( _stricmp(keyword,"insert") ==0 ) insertName  = value;
      else if ( _stricmp(keyword,"ins")    ==0 ) insertName  = value;
      else if ( _stricmp(keyword,"i")      ==0 ) insertName  = value;
      else if ( _stricmp(keyword,"length") ==0 ) length    = atol(value);
      else if ( _stricmp(keyword,"len")    ==0 ) length    = atol(value);
      else if ( _stricmp(keyword,"l")      ==0 ) length    = atol(value);
      else if ( _stricmp(keyword,"overlay")==0 ) overlayName = value;
      else if ( _stricmp(keyword,"over")   ==0 ) overlayName = value;
      else if ( _stricmp(keyword,"o")      ==0 ) overlayName = value;
      else if ( _stricmp(keyword,"recsize")==0 ) recsize   = atol(value);
      else if ( _stricmp(keyword,"lrecl")  ==0 ) recsize   = atol(value);
      else if ( _stricmp(keyword,"by")     ==0 ) recsize   = atol(value);
      else if ( _stricmp(keyword,"r")      ==0 ) recsize   = atol(value);
      else if ( _stricmp(keyword,"start")  ==0 ) start     = atol(value);
      else if ( _stricmp(keyword,"s")      ==0 ) start     = atol(value);
      else if ( _stricmp(keyword,"to")     ==0 ) toName    = value;
      else if ( _stricmp(keyword,"t")      ==0 ) toName    = value;
      else oops("unrecogisable or missing keyword");
      }
   } // end parse
void Splice::openFiles()
   {
   // open source and target files
   switch ( source )
      {
      case cut:
         sourceFile  = new SeqFile("CUT",cutName,_O_RDWR | _O_BINARY, _SH_DENYRW, _S_IREAD | _S_IWRITE);
         break;
      case copy:
         sourceFile  = new SeqFile("COPY",copyName,_O_RDONLY | _O_BINARY, _SH_DENYRW, _S_IREAD | _S_IWRITE);
         break;
      case unknown:
      default:
         oops("bug #1 in program");
      } // end switch
   switch ( target )
      {
      case none:
         targetFile = NULL;
         break;
      case to:
         targetFile  = new SeqFile("TO",toName,
                                   _O_WRONLY | _O_BINARY | _O_TRUNC | _O_CREAT, _SH_DENYRW,
                                   S_IREAD | S_IWRITE);
         break;
      case append:
         targetFile = new SeqFile("APPEND",appendName,_O_WRONLY | _O_BINARY, _SH_DENYRW, _S_IREAD | _S_IWRITE);
         break;
      case overlay:
         targetFile  = new SeqFile("OVERLAY",overlayName,_O_WRONLY | O_BINARY, _SH_DENYRW, _S_IREAD | _S_IWRITE);
         break;
      case insert:
         targetFile  = new SeqFile("INSERT",insertName,_O_RDWR | _O_BINARY, _SH_DENYRW, _S_IREAD | _S_IWRITE);
         break;
      default:
         oops("bug #2 in program.");
      } // end switch
   } // end openFiles
void Splice::analyse ()
   {
   start = start * recsize;
   length = length * recsize;
   at = at * recsize;
   // treat append like OVERLAY AT
   if ( target == append ) at = targetFile->size;
   // reign in regions that slop outside the sourceFile
   if ( start >= sourceFile->size ) length = 0;
   else if ( start+length > sourceFile->size ) length = sourceFile->size - start;
   // regions on the target may well slop outside TargetFile
   // so we don't reign them in. We pad with zeroes.
   } // end analyse
void Splice::work()
   {
   if ( target != none )
      {
      // calculate how big the target file is going to be
      // so we can allocate its space in one fell swoop.
      int newTargetSize = targetFile->size;
      if ( at > newTargetSize ) newTargetSize = at;
      if ( target == insert ) newTargetSize += length;
      if ( at + length > newTargetSize ) newTargetSize = at + length;
      if ( newTargetSize != targetFile->size )
         Bulk::resize(targetFile->handle, newTargetSize);
      }
   // pad with zeroes if we are overlaying or inserting past the end
   if ( target != none && (at > targetFile->size) )
      {
      Bulk::zap(targetFile->handle,
                targetFile->size, // old size
                at - targetFile->size,
                0);
      }
   // open up a hole in the target if needed.
   if ( target == insert )
      {
      Bulk::move(targetFile->handle,
                 at,
                 targetFile->size /* old size */ - at,
                 at + length);
      }
   // copy the contents of the region to the second file if needed.
   if ( target != none )
      {
      Bulk::move(sourceFile->handle,
                 start,
                 length,
                 targetFile->handle,
                 at);
      }
   // close up the hole in the source if needed.
   if ( source == cut )
      {
      // move data down onto the hole
      Bulk::move(sourceFile->handle,
                 start + length,
                 sourceFile->size - (start + length),
                 start);
      // shrink source file
      Bulk::resize(sourceFile->handle,
                   sourceFile->size - length);
      }
   } // end work
void Splice::closeFiles()
   {
   if ( sourceFile ) delete sourceFile;
   if ( targetFile ) delete targetFile;
   } // end closeFiles
void Splice::splice( int argc,  char *argv[] )
   {
   // master routine to handle the splice command line
   parse(argc,argv);
   checkParms();
   openFiles();
   analyse();
   work();
   closeFiles();
   }
// =============================================================================
void oops( char* reason )
   {
   banner();
   printf("Oops, error in SPLICE command line\n");
   printf(reason);
   printf("\ntry:\n"
          "SPLICE CUT  MyFile1.Txt START 99 LENGTH 99                           RECSIZE 1\n"
          "SPLICE CUT  MyFile1.Txt START 99 LENGTH 99 TO      MyFile2.txt       RECSIZE 1\n"
          "SPLICE CUT  MyFile1.Txt START 99 LENGTH 99 APPEND  MyFile2.txt       RECSIZE 1\n"
          "SPLICE CUT  MyFile1.Txt START 99 LENGTH 99 OVERLAY MyFile2.txt AT 99 RECSIZE 1\n"
          "SPLICE CUT  MyFile1.Txt START 99 LENGTH 99 INSERT  MyFile2.txt AT 99 RECSIZE 1\n"
          "SPLICE COPY MyFile1.Txt START 99 LENGTH 99 TO      MyFile2.txt       RECSIZE 1\n"
          "SPLICE COPY MyFile1.Txt START 99 LENGTH 99 APPEND  MyFile2.txt       RECSIZE 1\n"
          "SPLICE COPY MyFile1.Txt START 99 LENGTH 99 OVERLAY MyFile2.txt AT 99 RECSIZE 1\n"
          "SPLICE COPY MyFile1.Txt START 99 LENGTH 99 INSERT  MyFile2.txt AT 99 RECSIZE 1\n");
   exit (2);
   } // end oops
void banner(void)
   {
   /* display copyright banner, with shaded splashes */
   printf("\n" "\xb0\xb1\xb2\xdb" " Splice 1.5 "  "\xdb\xb2\xb1\xb0"
          "\n"
          "\nSplices a chunk of a file into another file."
          "\n"
          "\nCopyright: (c) 1997-2017 Canadian Mind Products"
          "\n" "#101 - 2536 Wark Street, Victoria, BC Canada V8T 4G8"
          "\n" "tel:(250) 361-9093    http://mindprod.com"
          "\n" "May be used freely for non-military use only"
          "\n\n");
   } /* end Banner */
/*****************/
int main( int argc,  char *argv[] )
   {
   Splice::splice( argc, argv );
   return 0;
   } // end main
// =============================================================================
